#pragma once

#include <cstdint>

template<class T>
class Point 
{
public:
	constexpr Point()noexcept
		:
		x( 0 ), 
		y( 0 )
	{}
	constexpr Point( const T X, const T Y )noexcept
		:
		x( X ), 
		y( Y )
	{}

public:
	T x;
	T y;
};

using PointD = Point<double>;
using PointF = Point<float>;
using PointI = Point<int>;
using PointL = Point<long>;
using PointU = Point<unsigned int>;


template<class T>
class Offset
{
public:
	constexpr Offset() noexcept
		:
		deltaX( 0 ), 
		deltaY( 0 )
	{}
	constexpr Offset( const T DeltaX, const T DeltaY )noexcept
		:
		deltaX( DeltaX ), 
		deltaY( DeltaY )
	{}
public:
	T deltaX, deltaY;
};

using OffsetD = Offset<double>;
using OffsetF = Offset<float>;
using OffsetI = Offset<int>;
using OffsetL = Offset<long>;
using OffsetU = Offset<unsigned int>;

template<class T>
constexpr Point<T> operator+( const Point<T>& Pt, const Offset<T>& Off )noexcept
{
	return Point<T>( Pt.x + Off.deltaX, Pt.y + Off.deltaY );
}

template<class T>
constexpr Point<T> operator-( const Point<T>& Pt, const Offset<T>& Off )noexcept
{
	return Point<T>( Pt.x - Off.deltaX, Pt.y - Off.deltaY );
}


template<class T>
class Extent
{
public:
	constexpr Extent()noexcept 
		:
		width( 0 ), 
		height( 0 )
	{}
	constexpr Extent( const T Width, const T Height )noexcept
		:
		width( Width ),
		height( Height )
	{}

	constexpr T Area()const noexcept
	{
		return width * height;
	}

public:
	T width, height;
};

using ExtentD = Extent<double>;
using ExtentF = Extent<float>;
using ExtentI = Extent<int>;
using ExtentL = Extent<long>;
using ExtentU = Extent<unsigned int>;


template<class T>
constexpr Extent<T> operator*( const Extent<T>& Ext, const T Scalar )noexcept
{
	return Extent<T>( Ext.width * Scalar, Ext.height * Scalar );
}

template<class T>
constexpr Extent<T> operator*( const T Scalar, const Extent<T>& Ext)noexcept
{
	return Ext * Scalar;
}

template<class T>
constexpr Point<T> operator+( const Point<T>& Pt, const Extent<T>& Ext )noexcept
{
	return Point<T>( Pt.x + Ext.width, Pt.y + Ext.height );
}

template<class T>
constexpr Point<T> operator+( const Extent<T>& Ext, const Point<T>& Pt)noexcept
{
	return Point<T>( Pt.x + Ext.width, Pt.y, +Ext.height );
}

template<class T>
constexpr Point<T> operator-( const Point<T>& Pt, const Extent<T>& Ext )noexcept
{
	return Point<T>( Pt.x - Ext.width, Pt.y - Ext.height );
}

template<class T>
constexpr Extent<T> operator-( const Point<T>& Pt1, const Point<T>& Pt2 )noexcept
{
	return Extent<T>( Pt1.x - Pt2.x, Pt1.y - Pt2.y );
}

template<class T>
constexpr Extent<T> operator/( const Extent<T>& Ext, const T Scalar )noexcept
{
	return Extent<T>( Ext.width / Scalar, Ext.height / Scalar );
}

template<class T>
constexpr Extent<T> operator/( const T Scalar, const Extent<T>& Ext)noexcept
{
	return Extent<T>( Ext.width / Scalar, Ext.height / Scalar );
}

template<>
constexpr Extent<float> operator/( const Extent<float>& Ext, const float Scalar )noexcept
{
	return Ext * ( 1.f / Scalar );
}

template<>
constexpr Extent<float> operator/( const float Scalar, const Extent<float>& Ext )noexcept
{
	return Ext * ( 1.f / Scalar );
}

template<>
constexpr Extent<double> operator/( const Extent<double>& Ext, const double Scalar )noexcept
{
	return Ext * ( 1. / Scalar );
}

template<>
constexpr Extent<double> operator/( const double Scalar, const Extent<double>& Ext )noexcept
{
	return Ext * ( 1. / Scalar );
}


template<class T>
class Rect 
{
public:
	constexpr Rect() :leftTop( 0, 0 ), rightBottom( 0, 0 )noexcept
	{}
	constexpr Rect( const Point<T>& LeftTop, const Point<T>& RightBottom )noexcept
		:
		leftTop( LeftTop ), rightBottom( RightBottom )
	{}
	constexpr Rect( const Point<T>& LeftTop, const Extent<T>& Ext )noexcept
		:
		leftTop( LeftTop ), rightBottom( LeftTop + Ext )
	{}

	constexpr T Width()const noexcept
	{
		return rightBottom.x - leftTop.x;
	}
	constexpr T Height()const noexcept
	{
		return rightBottom.y - leftTop.y;
	}
	constexpr Extent<T> Dimensions()const noexcept
	{
		return rightBottom - leftTop;
	}
	constexpr Point<T> Center()const noexcept
	{
		return leftTop + ( ( rightBottom - leftTop ) / static_cast< T >( 2 ) );
	}
	constexpr bool Contains( const Point<T>& Pt )const noexcept
	{
		return
			( Pt.x >= leftTop.x && Pt.x < rightBottom.x ) &&
			( Pt.y >= leftTop.y && Pt.y < rightBottom.y );
	}

	constexpr Rect TranslateTo( const Point<T>& Pt )const noexcept
	{
		return{ Pt, rightBottom - leftTop };
	}
	Rect& TranslateTo( const Point<T>& Pt ) noexcept
	{
		*this = Rect( Pt, rightBottom - leftTop );
		return *this;
	}
	constexpr Rect TranslateBy( const Offset<T>& Off )const noexcept
	{
		return{ leftTop + Off, rightBottom + Off };
	}
	Rect& TranslateBy( const Offset<T>& Off ) noexcept
	{
		*this = Rect( leftTop + Off, rightBottom + Off );
		return *this;
	}

	constexpr Rect Resize( const Extent<T>& Ext )const noexcept
	{
		return{ leftTop, leftTop + Ext };
	}
	constexpr Rect ResizeFromCenter( const Extent<T>& Ext )const noexcept
	{
		return{ 
			Center() - ( Ext / static_cast< T >( 2 ) ), 
			Center() + ( Ext / static_cast< T >( 2 ) ) 
		};
	}
public:
	Point<T> leftTop, rightBottom;
};

using RectD = Rect<double>;
using RectF = Rect<float>;
using RectI = Rect<int>;
using RectL = Rect<long>;
using RectU = Rect<unsigned int>;
